import enum
import datetime

from typing import Optional
from sqlalchemy import (
    Column,
    Integer,
    Float,
    DateTime,
    String,
    Enum,
    func,
    select,
    distinct,
)

from logdetective.server.database.base import Base, transaction


class EndpointType(enum.Enum):
    """Different analyze endpoints"""

    ANALYZE = "analyze_log"
    ANALYZE_STAGED = "analyze_log_staged"
    ANALYZE_STREAM = "analyze_log_stream"


class AnalyzeRequestMetrics(Base):
    """Store data related to received requests and given responses"""

    __tablename__ = "analyze_request_metrics"

    id = Column(Integer, primary_key=True)
    endpoint = Column(
        Enum(EndpointType),
        nullable=False,
        index=True,
        comment="The service endpoint that was called",
    )
    request_received_at = Column(
        DateTime,
        nullable=False,
        index=True,
        default=datetime.datetime.now(datetime.timezone.utc),
        comment="Timestamp when the request was received",
    )
    log_url = Column(
        String,
        nullable=False,
        index=False,
        comment="Log url for which analysis was requested",
    )
    response_sent_at = Column(
        DateTime, nullable=True, comment="Timestamp when the response was sent back"
    )
    response_length = Column(
        Integer, nullable=True, comment="Length of the response in chars"
    )
    response_certainty = Column(
        Float, nullable=True, comment="Certainty for generated response"
    )

    @classmethod
    def create(
        cls,
        endpoint: EndpointType,
        log_url: str,
        request_received_at: Optional[datetime.datetime] = None,
    ) -> int:
        """Create AnalyzeRequestMetrics new line
        with data related to a received request"""
        with transaction(commit=True) as session:
            metrics = AnalyzeRequestMetrics()
            metrics.endpoint = endpoint
            metrics.request_received_at = request_received_at or datetime.datetime.now(
                datetime.timezone.utc
            )
            metrics.log_url = log_url
            session.add(metrics)
            session.flush()
            return metrics.id

    @classmethod
    def update(
        cls,
        id_: int,
        response_sent_at: datetime,
        response_length: int,
        response_certainty: float,
    ) -> None:
        """Update an AnalyzeRequestMetrics line
        with data related to the given response"""
        with transaction(commit=True) as session:
            metrics = session.query(AnalyzeRequestMetrics).filter_by(id=id_).first()
            metrics.response_sent_at = response_sent_at
            metrics.response_length = response_length
            metrics.response_certainty = response_certainty
            session.add(metrics)

    @classmethod
    def _get_requests_by_time_for_postgres(
        cls, start_time, end_time, time_format, endpoint
    ):
        """func.to_char is PostgreSQL specific.
        Let's unit tests replace this function with the SQLite version.
        """
        if time_format == "%Y-%m-%d":
            pgsql_time_format = "YYYY-MM-DD"
        else:
            pgsql_time_format = "YYYY-MM-DD HH24"

        requests_by_time_format = (
            select(
                cls.id,
                func.to_char(cls.request_received_at, pgsql_time_format).label(
                    "time_format"
                ),
            )
            .filter(cls.request_received_at.between(start_time, end_time))
            .filter(cls.endpoint == endpoint)
            .cte("requests_by_time_format")
        )
        return requests_by_time_format

    @classmethod
    def _get_requests_by_time_for_sqllite(
        cls, start_time, end_time, time_format, endpoint
    ):
        """func.strftime is SQLite specific.
        Use this function in unit test using flexmock:

        flexmock(AnalyzeRequestMetrics).should_receive("_get_requests_by_time_for_postgres")
        .replace_with(AnalyzeRequestMetrics._get_requests_by_time_for_sqllite)
        """
        requests_by_time_format = (
            select(
                cls.id,
                func.strftime(time_format, cls.request_received_at).label(
                    "time_format"
                ),
            )
            .filter(cls.request_received_at.between(start_time, end_time))
            .filter(cls.endpoint == endpoint)
            .cte("requests_by_time_format")
        )
        return requests_by_time_format

    @classmethod
    def get_requests_in_period(
        cls,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        time_format: str,
        endpoint: Optional[EndpointType] = EndpointType.ANALYZE,
    ) -> dict[datetime.datetime, int]:
        """
        Get a dictionary with request counts grouped by time units within a specified period.

        Args:
            start_time (datetime): The start of the time period to query
            end_time (datetime): The end of the time period to query
            time_format (str): The strftime format string to format timestamps (e.g., '%Y-%m-%d')
            endpoint (EndpointType): The analyze API endpoint to query

        Returns:
            dict[datetime, int]: A dictionary mapping datetime objects to request counts
        """
        with transaction(commit=False) as session:
            requests_by_time_format = cls._get_requests_by_time_for_postgres(
                start_time, end_time, time_format, endpoint
            )

            count_requests_by_time_format = select(
                requests_by_time_format.c.time_format,
                func.count(distinct(requests_by_time_format.c.id)),  # pylint: disable=not-callable
            ).group_by("time_format")

            counts = session.execute(count_requests_by_time_format)
            results = counts.fetchall()

            # Convert results to a dictionary with proper datetime keys
            counts_dict = {
                datetime.datetime.strptime(r[0], time_format): r[1] for r in results
            }

            return counts_dict



import os
import logging
from typing import Tuple

import drain3
from drain3.template_miner_config import TemplateMinerConfig
from llama_cpp import Llama, LlamaGrammar

from logdetective.constants import SUMMARIZE_PROMPT_TEMPLATE
from logdetective.utils import get_chunks

LOG = logging.getLogger("logdetective")


class LLMExtractor:
    """
    A class that extracts relevant information from logs using a language model.
    """

    def __init__(self, model: Llama, n_lines: int = 2):
        self.model = model
        self.n_lines = n_lines
        self.grammar = LlamaGrammar.from_string(
            'root ::= ("Yes" | "No")', verbose=False
        )

    def __call__(
        self, log: str, n_lines: int = 2, neighbors: bool = False
    ) -> list[str]:
        chunks = self.rate_chunks(log)
        out = self.create_extract(chunks, neighbors)
        return out

    def rate_chunks(self, log: str) -> list[tuple]:
        """Scan log by the model and store results.

        :param log: log file content
        """
        results = []
        log_lines = log.split("\n")

        for i in range(0, len(log_lines), self.n_lines):
            block = "\n".join(log_lines[i: i + self.n_lines])
            prompt = SUMMARIZE_PROMPT_TEMPLATE.format(log)
            out = self.model(prompt, max_tokens=7, grammar=self.grammar)
            out = f"{out['choices'][0]['text']}\n"
            results.append((block, out))

        return results

    def create_extract(self, chunks: list[tuple], neighbors: bool = False) -> list[str]:
        """Extract interesting chunks from the model processing."""
        interesting = []
        summary = []
        # pylint: disable=consider-using-enumerate
        for i in range(len(chunks)):
            if chunks[i][1].startswith("Yes"):
                interesting.append(i)
                if neighbors:
                    interesting.extend([max(i - 1, 0), min(i + 1, len(chunks) - 1)])

        interesting = set(interesting)

        for i in interesting:
            summary.append(chunks[i][0])

        return summary


class DrainExtractor:
    """A class that extracts information from logs using a template miner algorithm."""

    def __init__(self, verbose: bool = False, context: bool = False, max_clusters=8):
        config = TemplateMinerConfig()
        config.load(f"{os.path.dirname(__file__)}/drain3.ini")
        config.profiling_enabled = verbose
        config.drain_max_clusters = max_clusters
        self.miner = drain3.TemplateMiner(config=config)
        self.verbose = verbose
        self.context = context

    def __call__(self, log: str) -> list[Tuple[int, str]]:
        out = []
        # First pass create clusters
        for _, chunk in get_chunks(log):
            processed_chunk = self.miner.add_log_message(chunk)
            LOG.debug(processed_chunk)
        # Sort found clusters by size, descending order
        sorted_clusters = sorted(
            self.miner.drain.clusters, key=lambda it: it.size, reverse=True
        )
        # Second pass, only matching lines with clusters,
        # to recover original text
        for chunk_start, chunk in get_chunks(log):
            cluster = self.miner.match(chunk, "always")
            if cluster in sorted_clusters:
                out.append((chunk_start, chunk))
                sorted_clusters.remove(cluster)
        return out
